{-|
Module      : Data.Graph.Clustering.HLouvain
Description : Purely functional (Inductive) Louvain clustering
Copyright   : (c) CNRS, 2020-Present
License     : AGPL + CECILL v3
Maintainer  : alexandre.delanoe+louvain@iscpif.fr
Stability   : experimental
Portability : POSIX


Main Reference:
 * Blondel, Vincent D; Guillaume, Jean-Loup; Lambiotte, Renaud; Lefebvre, Etienne (9 October 2008). "Fast unfolding of communities in large networks". Journal of Statistical Mechanics: Theory and Experiment. 2008 (10): P10008. arXiv:0803.0476 Freely accessible. doi:10.1088/1742-5468/2008/10/P10008.
 * Louvain_Modularity https://en.wikipedia.org/wiki/Louvain_Modularity
-}


module Data.Graph.Clustering.HLouvain 
  where

import Data.Ord (Down(..))
import Data.List (sortOn)
import Data.Graph.Inductive
-- import Data.Graph.Clustering.Louvain

------------------------------------------------------------------------
-- | Main Tools
------------------------------------------------------------------------

glustering :: DynGraph gr
           => gr a b
           -> [Module]
glustering = undefined

------------------------------------------------------------------------
-- | Definitions
------------------------------------------------------------------------

data Module = Module { node    :: Node
                     , modules :: [Module]
                     , score   :: Double
                     }

modulare :: DynGraph gr
         => gr a b
         ->  Module
         -> [Module]
         -> [Module]
modulare _ m []  = [m]
modulare _ m [n] = sortOn (Down . score) [m,n]
modulare g m0@(Module n ms0 _) m1@(m@(Module n1 _ _):ms)
  | score1 > score2 = [Module n1 (modulare g m0 ms0) score1] ++ ms
  | otherwise       = [m]  ++ (modulare g m0 ms)
    where
      score1 = sum $ map (modularity g) [m0, m]
      score2 = sum $ map (modularity g) ([m0]++ms)

flat :: Module -> [Node]
flat (Module n [] _) = [n]
flat (Module n ns _) = [n] ++ (concat $ map flat ns)

-- | TODO Sum or Union -> QuickCheck
modularity :: DynGraph gr => gr a b -> Module -> Double
modularity g m = modularity' g (flat m)
  where
    modularity' :: DynGraph gr => gr a b -> [Node] -> Double
    modularity' gr ns = coverage - edgeDensity
        where
            coverage :: Double
            coverage  = sizeSubGraph / sizeAllGraph
                where
                    sizeSubGraph :: Double
                    sizeSubGraph = fromIntegral ( size $ subgraph ns gr )

                    sizeAllGraph :: Double
                    sizeAllGraph = fromIntegral (size gr)

            edgeDensity :: Double
            edgeDensity = (sum (Prelude.map (\node -> (degree node) / links ) ns)) ** 2
                where
                    degree :: Node -> Double
                    degree node = fromIntegral (deg gr node)

                    links :: Double
                    links       = fromIntegral (2 * (size gr))

